-----------------------------------------------------------------------
-- Upvalued Lua API.
-----------------------------------------------------------------------
local _G = getfenv(0)

-- Functions
local next = _G.next
local pairs = _G.pairs
local time = _G.time
local tonumber = _G.tonumber
local type = _G.type

-- Libraries
local math = _G.math
local table = _G.table

-----------------------------------------------------------------------
-- AddOn namespace.
-----------------------------------------------------------------------
local ADDON_NAME, private = ...

local LibStub = _G.LibStub

local BTC = LibStub("AceAddon-3.0"):NewAddon(ADDON_NAME, "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME)

local LQT = LibStub:GetLibrary("LibQTip-1.0")
local LDBIcon = LibStub("LibDBIcon-1.0")

local ADDON_NAME_SHORT = "BTC"
_G[ADDON_NAME_SHORT] = BTC

local addon = BTC

local addonversion
do
	local debug_version = false
	local alpha_version = false

	--[===[@debug@
	debug_version = true
	--@end-debug@]===]

	--@alpha@
	alpha_version = true
	--@end-alpha@

	local addon_version = _G.GetAddOnMetadata(ADDON_NAME, "Version")
	addonversion = debug_version and "Devel" or (alpha_version and addon_version .. "-Alpha") or addon_version
end -- do-block
local addonauthor = "James D. Callahan III"
local builddate = "20121025055459"

------------------------------------------------------------------------------
-- Constants.
------------------------------------------------------------------------------
local SPECIAL_CASE_COOLDOWNS = {
	-- Snowmaster 9000
	[21935] = {
		name = L["Snowmaster 9000"],
		duration = 72000,
	},
	-- Create Elune Stone
	[26265] = {
		name = L["Elune's Lantern"],
		duration = 72000,
	},
}

local COOLDOWN_SPELL_MAP = {
	-- Initialize my cooldowns list
	-- Alchemy
	-- Northrend Alchemy Research
	[60893] = true, -- Northrend Alchemy Research
	-- Metals (shared cooldown)
	[11479] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Iron to Gold
	[11480] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Mithril to Truesilver
	-- rm'd in 3.3.0
	-- [60350] = _G.STRING_SCHOOL_ELEMENTAL,		-- Transmute: Titanium
	-- Elements (shared cooldown)
	[17559] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Air to Fire
	[17566] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Earth to Life
	[17561] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Earth to Water
	--[[ No cooldown on Elemental Fire
			 [25146] = _G.STRING_SCHOOL_ELEMENTAL,
			 ]] --
	[17560] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Fire to Earth
	[17565] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Life to Earth
	[17563] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Undeath to Water
	[17562] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Water to Air
	[17564] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Water to Undeath
	-- Primals
	[28566] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Air to Fire
	[28585] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Earth to Life
	[28567] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Earth to Water
	[28568] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Fire to Earth
	[28583] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Fire to Mana
	[28584] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Life to Earth
	[28582] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Mana to Fire
	[28580] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Shadow to Water
	[28569] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Water to Air
	[28581] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Water to Shadow
	-- Eternals
	[53777] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Air to Earth
	[52776] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Air to Water
	[53781] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Earth to Air
	[53782] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Earth to Shadow
	[53775] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Fire to Life
	[53774] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Fire to Water
	[53773] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Life to Fire
	[53771] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Life to Shadow
	[54020] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Might
	[53779] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Shadow to Earth
	[52780] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Shadow to Life
	[53783] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Water to Air
	[53784] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Water to Fire
	-- Epic gems
	[66658] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : Ametrine
	[66659] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : Cardinal Ruby
	[66660] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : King's Amber
	[66662] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : Dreadstone
	[66663] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : Majestic Zircon
	[66664] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : Eye of Zul
	-- Diamonds (shared cooldown)
	--[[ It appears these were removed as well
			 [32765] = L["Diamonds"],		-- Transmute: Earthstorm Diamond
			 [32766] = L["Diamonds"],		-- Transmute: Skyfire Diamond
			 [57425] = L["Diamonds"],		-- Transmute: Skyflare Diamond
			 [57427] = L["Diamonds"],		-- Transmute: Earthsiege Diamond
			 ]] --
	-- New Cata Elemental transmutes
	[78866] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Living Elements
	[80243] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Truegold
	[80244] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Pyrium Bar
	-- MoP
	[114780] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Living Steel

	-- New Cata Gems
	--[[ Removed somewhere around 4.1
			 [80245] = L["Cataclysm Blue Gems"],		-- Transmute: Inferno Ruby
			 [80246] = L["Cataclysm Blue Gems"],		-- Transmute: Ocean Sapphire
			 [80247] = L["Cataclysm Blue Gems"],		-- Transmute: Amberjewel
			 [80248] = L["Cataclysm Blue Gems"],		-- Transmute: Demonseye
			 [80250] = L["Cataclysm Blue Gems"],		-- Transmute: Ember Topaz
			 [80251] = L["Cataclysm Blue Gems"],		-- Transmute: Dream Emerald
			 ]] --

	-- Enchanting
	[28027] = true, -- Prismatic Sphere
	[28028] = true, -- Void Sphere
	[116499] = true, -- Sha Crystal

	-- Jewelcrafting
	[47280] = true, -- Brilliant Glass
	[62242] = true, -- Icy Prism
	[73478] = true, -- Fire Prism

	-- MoP Jewelcrafting (shared cooldown)
	[131691] = L["Facets of Research"], -- Imperial Amethyst
	[131686] = L["Facets of Research"], -- Primordial Ruby
	[131593] = L["Facets of Research"], -- River's Heart
	[131695] = L["Facets of Research"], -- Sun's Radiance
	[131690] = L["Facets of Research"], -- Vermilion Onyx
	[131688] = L["Facets of Research"], -- Wild Jade

	-- Tailoring
	-- Wrath of the Lich King
	[56005] = true, -- Glacial Bag
	-- Cataclysm
	[75141] = true, -- Dream of Skywall
	[75145] = true, -- Dream of Ragnaros
	[75144] = true, -- Dream of Hyjal
	[75142] = true, -- Dream of Deepholm
	[75146] = true, -- Dream of Azshara
	-- Mists of Pandaria
	[125557] = true, -- Imperial Silk
	-- Inscription
	[61288] = true, -- Minor Inscription Research
	[61177] = true, -- Northrend Inscription Research
	[86654] = true, -- Horde Forged Documents
	[89244] = true, -- Alliance Forged Documents
	[112996] = true, -- Scroll of Wisdom
}


------------------------------------------------------------------------------
-- Variables.
------------------------------------------------------------------------------
local db
local tooltip

local primary = {}
local secondary = {}


------------------------------------------------------------------------------
-- Helper functions.
------------------------------------------------------------------------------
do
	local CHAT_MESSAGE_FORMAT = "\124TInterface\\Addons\\" .. ADDON_NAME .. "\\icon:" .. "%d" .. "\124t" .. ADDON_NAME_SHORT .. ": " .. "%s"

	function addon.Print(message)
		if not message then
			return
		end
		_G.DEFAULT_CHAT_FRAME:AddMessage(CHAT_MESSAGE_FORMAT:format(math.ceil(_G.select(2, _G.DEFAULT_CHAT_FRAME:GetFont())), message), 0.6, 1.0, 1.0)
	end
end -- do-block

function addon.Debug(message)
	if not message or not db.options.debug then
		return
	end
	addon.Print("DEBUG: " .. message)
end

-- coloring functions for tooltip text
local function Red(text) return "|cffff0000" .. text .. "|r" end

local function Orange(text) return "|cffdea51c" .. text .. "|r" end

local function Yellow(text) return "|cffffff00" .. text .. "|r" end

local function Green(text) return "|cff00ff00" .. text .. "|r" end

local function White(text) return "|cffffffff" .. text .. "|r" end

local function BTC_prettyTime(seconds)
	local remaining = seconds
	local output = ""

	if remaining >= 86400 then
		output = output .. math.floor(remaining / 86400) .. "d"
		remaining = remaining % 86400
	end

	if remaining >= 3600 then
		output = output .. math.floor(remaining / 3600) .. "h"
		remaining = remaining % 3600
	end

	if remaining >= 60 then
		output = output .. math.floor(remaining / 60) .. "m"
	end
	return output
end

local function PercentColorText(min, max, text)
	local num = min + 0
	local den = max + 0

	if den == 0 then
		return Green(text)
	end
	local perc = num / den
	addon.Debug(min .. "/" .. max .. " = " .. perc)

	if perc < 0.25 then
		return Red(text)
	elseif perc < 0.5 then
		return Orange(text)
	elseif perc < .75 then
		return Yellow(text)
	else
		return Green(text)
	end
end

do
	local header_list = {}
	local skill_frame_shown

	local function SearchForCooldowns()
		addon.Debug("in SearchForCooldowns")

		if not skill_frame_shown then
			return
		end
		------------------------------------------------------------------------------
		-- Save the current state of the TradeSkillFrame so it can be restored after we muck with it.
		------------------------------------------------------------------------------
		local have_materials = _G.TradeSkillFrame.filterTbl.hasMaterials
		local have_skillup = _G.TradeSkillFrame.filterTbl.hasSkillUp

		if have_materials then
			_G.TradeSkillFrame.filterTbl.hasMaterials = false
			_G.TradeSkillOnlyShowMakeable(false)
		end

		if have_skillup then
			_G.TradeSkillFrame.filterTbl.hasSkillUp = false
			_G.TradeSkillOnlyShowSkillUps(false)
		end
		_G.SetTradeSkillInvSlotFilter(0, 1, 1)
		_G.TradeSkillUpdateFilterBar()
		_G.TradeSkillFrame_Update()

		------------------------------------------------------------------------------
		-- Expand headers and scan for cooldowns.
		------------------------------------------------------------------------------
		table.wipe(header_list)

		for skill_index = _G.GetNumTradeSkills(), 1, -1 do
			local entry_name, entry_type, _, is_expanded = _G.GetTradeSkillInfo(skill_index)

			if (entry_type == "header" or entry_type == "subheader") and not is_expanded then
				header_list[entry_name] = true
				_G.ExpandTradeSkillSubClass(skill_index)
			end
		end

		for skill_index = 1, _G.GetNumTradeSkills() do
			local entry_name, entry_type = _G.GetTradeSkillInfo(skill_index)

			if entry_type ~= "header" and entry_type ~= "subheader" then
				local spell_id = tonumber(_G.GetTradeSkillRecipeLink(skill_index):match("^|c%x%x%x%x%x%x%x%x|H%w+:(%d+)"))

				if COOLDOWN_SPELL_MAP[spell_id] then
					local cooldown = _G.GetTradeSkillCooldown(skill_index)
					local cooldown_entry = db.cooldowns[spell_id]

					if not cooldown_entry then
						cooldown_entry = {}
						db.cooldowns[spell_id] = cooldown_entry
					end
					addon.Debug(entry_name .. ", cd is " .. _G.tostring(cooldown))

					local mapped_name = COOLDOWN_SPELL_MAP[spell_id]

					if type(mapped_name) == "string" then
						cooldown_entry.displayName = mapped_name
					else
						cooldown_entry.displayName = entry_name
					end

					if cooldown then
						cooldown_entry.endtime = time() + math.floor(cooldown)
					else
						cooldown_entry.endtime = 0
					end
				else
					-- Just in case we've stored an entry for a spell that was removed from the game.
					db.cooldowns[spell_id] = nil
				end
			end
		end
		------------------------------------------------------------------------------
		-- Restore the state of the things we changed.
		------------------------------------------------------------------------------
		for skill_index = _G.GetNumTradeSkills(), 1, -1 do
			local entry_name, _, _, is_expanded = _G.GetTradeSkillInfo(skill_index)

			if header_list[entry_name] then
				_G.CollapseTradeSkillSubClass(skill_index)
			end
		end
		_G.TradeSkillFrame.filterTbl.hasMaterials = have_materials
		_G.TradeSkillOnlyShowMakeable(have_materials)
		_G.TradeSkillFrame.filterTbl.hasSkillUp = have_skillup
		_G.TradeSkillOnlyShowSkillUps(have_skillup)

		_G.TradeSkillUpdateFilterBar()
		_G.TradeSkillFrame_Update()
	end

	-- I don't actually care about what updated. Just that it did. I only turn on monitoring
	-- for this event when a SPELL_CAST_START has happened for a spell (tradeskill) that has a
	-- cooldown has occurred. And only when it isn't an enchant.
	function BTC:BAG_UPDATE(...)
		addon.Debug("BAG_UPDATE")
		BTC:UnregisterEvent("BAG_UPDATE")

		if skill_frame_shown then
			addon.Debug("SearchForCooldowns")
			SearchForCooldowns()
		elseif not db.options.hideReminders then
			addon.Print(L["CLOSED_TOO_EARLY"])
		end
		addon.updateText()
	end

	function BTC:TRADE_SKILL_SHOW(eventName, ...)
		skill_frame_shown = true
		addon.Debug("In " .. eventName)

		if _G.IsTradeSkillLinked() or _G.IsTradeSkillGuild() then
			addon.Debug("Someone else's tradeskill")
			return
		end
		SearchForCooldowns()
		addon.updateText()
	end

	function BTC:TRADE_SKILL_CLOSE(...)
		skill_frame_shown = false
	end
end -- do-block

function BTC:COMBAT_LOG_EVENT_UNFILTERED(_, _, sub_event, _, _, source_name, _, _, _, _, _, _, spell_id)
	if sub_event ~= "SPELL_CAST_START" or source_name ~= BTC.me then
		return
	end
	local now = time()

	local special_case = SPECIAL_CASE_COOLDOWNS[spell_id]

	if special_case then
		local entry = db.cooldowns[spell_id]

		if not entry then
			entry = {}
			db.cooldowns[spell_id] = entry
		end
		entry.displayName = special_case.name
		entry.endtime = now + special_case.duration
	elseif COOLDOWN_SPELL_MAP[spell_id] then
		if not db.options.hideReminders then
			addon.Print(L["COOLDOWN_DETECTED"])
		end
		-- Right now, all spells with cooldowns place a new item into your bags. Right at the second this
		-- code executes, the player will still be in mid-spell-cast of the spell in question. So in order
		-- to determine when to check for a new cooldown, I'm now going to register for a BAG_UPDATE (the
		-- item hitting my bags), and then scan for new skills.
		BTC:RegisterEvent("BAG_UPDATE")
	else
		addon.Debug("Detected spellcast that wasn't in list. spellid = " .. spell_id)
	end
end

local function UpdateSkills()
	primary[1], primary[2], secondary[1], secondary[2], secondary[3], secondary[4] = _G.GetProfessions()

	for index = 1, 2 do
		if primary[index] then
			local name, _, rank, max_rank = _G.GetProfessionInfo(primary[index])
			db.primary[name] = db.primary[name] or {}
			db.primary[name].currlevel = rank
			db.primary[name].maxlevel = max_rank
		end
	end

	for index = 1, 4 do
		if secondary[index] then
			local name, _, rank, max_rank = _G.GetProfessionInfo(secondary[index])
			db.secondary[name] = db.secondary[name] or {}
			db.secondary[name].currlevel = rank
			db.secondary[name].maxlevel = max_rank
		end
	end
end

local secure_frame = _G.CreateFrame("Button", ADDON_NAME .. "SecureFrame", _G.UIParent, "SecureActionButtonTemplate")
secure_frame:SetAttribute("type", "macro")
secure_frame:SetFrameStrata("TOOLTIP")
secure_frame:Hide()

secure_frame:SetScript("OnLeave", function(self)
	self:SetParent(nil)
	self:ClearAllPoints()
	self:Hide()
end)

local highlight_texture = secure_frame:CreateTexture(nil, "OVERLAY")
highlight_texture:SetTexture([[Interface\QuestFrame\UI-QuestTitleHighlight]])
highlight_texture:SetBlendMode("ADD")
highlight_texture:SetAllPoints(secure_frame)

local function Profession_OnEnter(cell, profession_name)
	secure_frame:SetParent(cell)
	secure_frame:SetPoint("TOPLEFT", cell, "TOPLEFT")
	secure_frame:SetPoint("BOTTOMRIGHT", cell, "BOTTOMRIGHT")
	secure_frame:SetAttribute("macrotext", ("/cast %s\n"):format(profession_name))
	secure_frame:Show()
end

local DrawTooltip
local function Tooltip_OnMouseDown(cell, arg, button)
	local entry_type, toon_name, spell_id = (":"):split(arg)

	spell_id = tonumber(spell_id)

	if _G.IsAltKeyDown() then
		local cooldowns = BTC.db.factionrealm[toon_name].cooldowns

		if entry_type == "name" then
			if button == "LeftButton" then
				for k, v in pairs(cooldowns) do
					cooldowns[k] = nil
				end
			elseif button == "RightButton" then
				for k, v in pairs(cooldowns) do
					v.hide = nil
				end
			end
		elseif entry_type == "spell" then
			if button == "LeftButton" then
				cooldowns[spell_id] = nil
			elseif button == "RightButton" and cooldowns[spell_id] then
				cooldowns[spell_id].hide = true
			end
		end
		DrawTooltip()
	end
	addon.updateText()
end

local GREEN = "|cff00ff00"
local RED = "|cffff0000"
local YELLOW = "|cffffff00"

local function Tooltip_AddProfessionLine(profession_name, data)
	if data.currlevel == data.maxlevel then
		return tooltip:AddLine(Yellow(("   %s"):format(profession_name)), " ", Yellow(("(%d/%d)"):format(data.currlevel, data.maxlevel)))
	else
		return tooltip:AddLine(Yellow(("   %s"):format(profession_name)), " ", ("%s(%s%d%s/%s%d%s)"):format(YELLOW, RED, data.currlevel, YELLOW, GREEN, data.maxlevel, YELLOW))
	end
end

local spell_registry = {}

 function DrawTooltip()
	local line_num = 0
	local handler_arg = ""
	local now = time()

	tooltip:Clear()
	tooltip:AddHeader(White(_G.PRIMARY_SKILLS))

	for profession_name, data in pairs(db.primary) do
		line_num = Tooltip_AddProfessionLine(profession_name, data)
		tooltip:SetLineScript(line_num, "OnEnter", Profession_OnEnter, profession_name)
	end

	if db.options.showSecondary then
		tooltip:AddLine(" ")
		tooltip:AddHeader(White(_G.SECONDARY_SKILLS))

		for profession_name, data in pairs(db.secondary) do
			line_num = Tooltip_AddProfessionLine(profession_name, data)
			tooltip:SetLineScript(line_num, "OnEnter", Profession_OnEnter, profession_name)
		end
	end
	tooltip:AddLine(" ")
	tooltip:AddHeader(White(L["Cooldowns :"]))

	for toon_name, data in pairs(BTC.db.factionrealm) do
		if next(data.cooldowns) then
			line_num = tooltip:AddLine(Yellow("   " .. toon_name))
			tooltip:SetLineScript(line_num, "OnMouseDown", Tooltip_OnMouseDown, "name:" .. toon_name)
		end
		table.wipe(spell_registry)

		for spell_id, data in pairs(data.cooldowns) do
			local display_name = data.displayName

			if not data.hide and not spell_registry[display_name] then
				if data.endtime == 0 or now > data.endtime then
					line_num = tooltip:AddLine(Yellow("      " .. display_name), " ", Green(L["Ready"]))
				else
					line_num = tooltip:AddLine(Yellow("      " .. display_name), " ", Red(BTC_prettyTime(data.endtime - now)))
				end
				tooltip:SetLineScript(line_num, "OnMouseDown", Tooltip_OnMouseDown, "spell:" .. toon_name .. ":" .. spell_id)
				spell_registry[display_name] = true
			end
		end
	end
	tooltip:AddLine(" ")
	line_num = tooltip:AddLine("b")
	tooltip:SetCell(line_num, 1, L["Click for Configuration and Help"], "LEFT", 3)
end

local LDB = LibStub("LibDataBroker-1.1", true)
local launcher = LDB:NewDataObject("BTC", {
	type = "data source",
	text = "",
	label = "Broker_TradeCooldowns",
	icon = "Interface\\Addons\\Broker_TradeCooldowns\\icon",
	OnClick = function(clickedframe, button)
		BTC:ShowConfig()
		addon.updateText()
	end,
	OnEnter = function(self)
		tooltip = LQT:Acquire(ADDON_NAME_SHORT .. "TT", 3, "LEFT", "CENTER", "RIGHT") -- or GameTooltip
		tooltip:SetScale(db.options.tipscale)
		DrawTooltip()
		tooltip:SetAutoHideDelay(.1, self)
		tooltip:EnableMouse()
		tooltip:SmartAnchorTo(self)
		tooltip:Show()
	end,
	OnLeave = function()
	-- this should get handled by the autohide delay stuff now
	end,
})

-- options stuff
local displayOptions = nil
local function giveDisplay()
	if (not displayOptions) then
		displayOptions = {
			order = 4,
			type = "group",
			name = L["Display Options"],
			desc = L["DISPLAY_OPTIONS_DESC"],
			args = {
				header1 = {
					order = 41,
					type = "header",
					name = L["Display Options"],
				},
				desc = {
					order = 42,
					type = "description",
					name = L["DISPLAY_OPTIONS_DESC"],
				},
				tooltipscale = {
					order = 43,
					type = "range",
					name = L["Tooltip Scale"],
					desc = L["TOOLTIP_SCALE_DESC"],
					min = .7,
					max = 1.3,
					step = .05,
					bigStep = .05,
					get = function() return db.options.tipscale end,
					set = function(info, v)
						db.options.tipscale = v
						tooltip:SetScale(v)
					end,
				},
				debug = {
					order = 44,
					type = "toggle",
					name = L["Debug"],
					desc = L["DEBUG_DESC"],
					get = function() return db.options.debug end,
					set = function() db.options.debug = not db.options.debug end,
				},
				showsecondary = {
					order = 46,
					type = "toggle",
					name = L["Show Secondary"],
					desc = L["SHOW_SECONDARY_DESC"],
					get = function() return db.options.showSecondary end,
					set = function() db.options.showSecondary = not db.options.showSecondary end,
				},
				hidereminders = {
					order = 47,
					type = "toggle",
					name = L["Hide Reminders"],
					desc = L["HIDE_REMINDERS_DESC"],
					get = function() return db.options.hideReminders end,
					set = function() db.options.hideReminders = not db.options.hideReminders end,
				},
				header2 = {
					order = 200,
					type = "header",
					name = L["Minimap Icon Options"],
				},
				desc2 = {
					order = 201,
					type = "description",
					name = L["MINIMAP_OPTIONS_DESC"],
				},
				minimap_icon = {
					order = 202,
					type = "toggle",
					width = "full",
					name = L["Show Minimap Icon"],
					desc = L["Draws the icon on the minimap."],
					get = function()
						return not db.options.minimap_icon.hide
					end,
					set = function(info, value)
						db.options.minimap_icon.hide = not value
						LDBIcon[value and "Show" or "Hide"](LDBIcon, ADDON_NAME_SHORT)
					end,
				},
				header3 = {
					order = 301,
					type = "header",
					name = L["Help and Usage Notes"],
				},
				help1 = {
					order = 302,
					type = "description",
					name = "- " .. L["HELP_STRING_0"],
				},
				help2 = {
					order = 303,
					type = "description",
					name = "- " .. L["HELP_STRING_1"],
				},
				help3 = {
					order = 304,
					type = "description",
					name = "- " .. L["HELP_STRING_2"],
				},
				help4 = {
					order = 305,
					type = "description",
					name = "- " .. L["HELP_STRING_3"],
				},
				help5 = {
					order = 306,
					type = "description",
					name = "- " .. L["HELP_STRING_4"],
				},
			},
		}
	end
	return displayOptions
end

local textOptions = nil
local function giveText()
	if (not textOptions) then
		textOptions = {
			order = 4,
			type = "group",
			name = L["Text Options"],
			desc = L["TEXT_OPTIONS_DESC"],
			args = {
				header1 = {
					order = 41,
					type = "header",
					name = L["Text Options"],
				},
				desc = {
					order = 42,
					type = "description",
					name = L["TEXT_OPTIONS_DESC"],
				},
				showtext = {
					order = 50,
					type = "toggle",
					name = L["Show Text"],
					desc = L["SHOW_TEXT_DESC"],
					get = function() return db.options.showText end,
					set = function()
						db.options.showText = not db.options.showText
						addon.updateText()
					end,
				},
				showcounters = {
					order = 51,
					type = "toggle",
					name = L["Show Counters"],
					desc = L["SHOW_COUNTERS_DESC"],
					get = function() return db.options.showCounters end,
					set = function()
						db.options.showCounters = not db.options.showCounters
						addon.updateText()
					end,
					disabled = function()
						if (db.options.showText) then
							return false
						else
							return true
						end
					end,
				},
				textcooldown = {
					order = 52,
					type = "select",
					name = L["Select Cooldown to Display"],
					desc = L["SHOW_TEXTCD_DESC"],
					get = function() return db.options.textCooldown end,
					set = function(info, value)
						db.options.textCooldown = value
						addon.updateText()
					end,
					values = function()
						local dbf = BTC.db.factionrealm
						local t = {}
						local index = ""
						for k, v in pairs(dbf) do
							local Displayed = {}
							for i, j in pairs(v.cooldowns) do
								if (Displayed[i] ~= false) then
									if (j.hide ~= nil) then
										Displayed[i] = false
									else
										if (Displayed[i] == true) then
											-- already displayed, don't do anything
										else
											index = k .. ":" .. i
											t[index] = k .. ":" .. j.displayName
											Displayed[i] = true
										end
									end
								end
							end
						end
						return t
					end,
					disabled = function()
						if (db.options.showText) then
							if (db.options.showCounters) then
								return true
							else
								return false
								--[[
								if ( next( db.cooldowns ) ~= nil ) then
									return false
								else
									return true
								end ]] --
							end
						else
							return true
						end
					end,
				},
			},
		}
	end
	return textOptions
end

local modularOptions = {}
local options = nil
local function fullOptions()
	if (not options) then
		options = {
			type = "group",
			name = ADDON_NAME_SHORT,
			args = {
				general = {
					order = 1,
					type = "group",
					name = L["General Information"],
					desc = L["GENERAL_INFO_DESC"],
					args = {
						header1 = {
							order = 11,
							type = "header",
							name = L["General Information"],
						},
						desc = {
							order = 12,
							type = "description",
							name = L["GENERAL_INFO_DESC"],
						},
						author = {
							order = 13,
							type = "description",
							name = L["Author : "] .. addonauthor .. "\n",
						},
						version = {
							order = 14,
							type = "description",
							name = L["Version : "] .. addonversion .. "\n",
						},
						builddate = {
							order = 15,
							type = "description",
							name = L["Build Date : "] .. builddate .. "\n",
						},
						dbversion = {
							order = 15,
							type = "description",
							name = L["Database Version : "] .. db.options.dbVersion .. "\n",
						},
						translators = {
							order = 16,
							type = "description",
							name = string.format(L["Helpful Translators (thank you) : %s"],
								"whynona2(frFR), SunTsu(deDE), Dessa(deDE), " ..
									"amunra(koKR), xska(ruRU), <CWDG>tnt2ray(zhCN), " ..
									"NightOw1(zhTW), andy52005(zhTW)\n")
						},
					},
				},
			},
		}
	end

	for k, v in pairs(modularOptions) do
		options.args[k] = (type(v) == "function") and v() or v
	end

	return options
end

local AceConfig = LibStub("AceConfig-3.0")
local AceConfigReg = LibStub("AceConfigRegistry-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

function addon:SetupOptions()
	AceConfigReg:RegisterOptionsTable(ADDON_NAME_SHORT, fullOptions)
	self.optionsFrame = AceConfigDialog:AddToBlizOptions(ADDON_NAME_SHORT, nil, nil, "general")

	-- Fill up our modular options...
	self:RegisterModuleOptions("Display", giveDisplay(), L["Display Options"])
	self:RegisterModuleOptions("Text", giveText(), L["Text Options"])
end

function addon:RegisterModuleOptions(name, optionsTable, displayName)
	modularOptions[name] = optionsTable
	self.optionsFrame[name] = AceConfigDialog:AddToBlizOptions(ADDON_NAME_SHORT, displayName, ADDON_NAME_SHORT, name)
end

function BTC:ShowConfig()
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrame["Display"])
end

function BTC.updateText()
	addon.Debug("In updateText")
	local newStr = " "
	if (db.options.showText) then
		if (db.options.showCounters) then
			-- show counter, like (8/10). Color text based on percent of CDs available
			local numCDs = 0
			local numAvail = 0
			local dbf = BTC.db.factionrealm
			for k, v in pairs(dbf) do
				local t = {}
				local Displayed = {}
				local Hidden = {}
				for i, a in pairs(v.cooldowns) do
					local blort = a.displayName
					if (a.hide ~= nil) then
						Hidden[blort] = true
					end
					if not (Displayed[blort]) then
						Displayed[blort] = true
						local now = time()
						local timeleft = a.starttime
						if ((a.endtime == 0) or (now > a.endtime)) then
							t[blort] = true
						else
							t[blort] = false
						end
					end
				end
				for i, a in pairs(t) do
					if not (Hidden[i]) then
						numCDs = numCDs + 1
						if (a == true) then
							numAvail = numAvail + 1
						end
					end
				end
			end
			newStr = "(" .. numAvail .. "/" .. numCDs .. ")"
			newStr = PercentColorText(numAvail, numCDs, newStr)
			addon.Debug("Counters = " .. newStr)
		else
			-- Show an individual cooldown, not a counter
			if (db.options.textCooldown ~= 0) then
				local dbf = addon.db.factionrealm
				-- Should be in the form: toonName:CooldownSpellID
				addon.Debug("textCooldown = " .. db.options.textCooldown)
				local selToon, selCD = (":"):split(db.options.textCooldown)
				if ((selToon == nil) or (selToon == "") or (selCD == nil)) then
					-- for some reason, its possible to have a blank toon here. No idea why. THrow an error.
					addon.Print(L["Couldn't assign value. Please file a bug ticket."])
					addon.Print(L["Include the following in the ticket:"])
					if (selToon == nil) then addon.Print("selToon is nil") end
					if (selToon == "") then addon.Print("selToon is blank") end
					if (selCD == nil) then addon.Print("selCD is nil") end
				else
					addon.Debug("toon = " .. selToon .. ", cd = " .. selCD)
					for k, v in pairs(dbf) do
						addon.Debug("k=" .. k)
						if (k == selToon) then
							addon.Debug("matched toon")
							for i, j in pairs(v.cooldowns) do
								addon.Debug("i=" .. i)
								selCD = selCD + 0
								if (i == selCD) then
									addon.Debug("matched spell")
									newStr = Yellow(j.displayName) .. ":"
									local now = time()
									if ((j.endtime == 0) or (now > j.endtime)) then
										newStr = newStr .. Green(L["Ready"])
									else
										local pretty = BTC_prettyTime(j.endtime - now)
										newStr = newStr .. Red(pretty)
									end
								end
							end
						end
					end
				end
			end
		end
	end
	launcher.text = newStr
end

function BTC_upgradeDB()
	local olddb = BTC.db.factionrealm[BTC.me]
	-- This checks the current DB version the person has, and upgrades as necessary
	if not (olddb.options) then
		-- They have the original version
		olddb.options = {}
		olddb.options = {
			dbVersion = "1.0",
			debug = false,
			showWeapons = false,
			showSecondary = false,
			showClass = false,
			showArmor = false,
			showLanguages = false,
		}
		olddb.weapons = {}
		olddb.primary = {}
		olddb.secondary = {}
		olddb.class = {}
		olddb.languages = {}
		olddb.armor = {}
		olddb.skills = nil
		addon.Print(string.format(L["Database upgraded to %s"], "1.0"))
	end
	if (olddb.options.dbVersion == "1.0") then
		olddb.options.tipscale = 1
		olddb.options.dbVersion = "1.1"
		addon.Print(string.format(L["Database upgraded to %s"], "1.1"))
	end
	if (olddb.options.dbVersion == "1.1") then
		olddb.options.showText = false
		olddb.options.textCooldown = 0
		olddb.options.dbVersion = "1.2"
		addon.Print(string.format(L["Database upgraded to %s"], "1.2"))
	end
	if (olddb.options.dbVersion == "1.2") then
		olddb.options.showClass = nil
		olddb.options.showArmor = nil
		olddb.options.showLanguages = nil
		olddb.languages = nil
		olddb.class = nil
		olddb.armor = nil
		olddb.options.dbVersion = "1.3"
		addon.Print(string.format(L["Database upgraded to %s"], "1.2"))
	end
	-- pretty much want to do this no matter what...
	local OldCleared = false
	for k, v in pairs(olddb.cooldowns) do
		if (COOLDOWN_SPELL_MAP[k] == nil) then
			olddb.cooldowns[k] = nil
			OldCleared = true
		end
	end
	if (OldCleared == true) then
		addon.Print(L["Cleared out old cooldowns."])
	end
	-------------------------------------------------
	if ((olddb.options.dbVersion == "1.3") or (olddb.options.dbVersion == "1.4")) then
		olddb.options.hideReminders = false
		olddb.options.dbVersion = "1.5"
		addon.Print(string.format(L["Database upgraded to %s"], "1.5"))
	end
	if (olddb.options.dbVersion == "1.5") then
		olddb.weapons = nil
		olddb.secondary[L["Riding"]] = nil
		olddb.options.showWeapons = nil
		olddb.options.dbVersion = "1.6"
		addon.Print(string.format(L["Database upgraded to %s"], "1.6"))
	end
	if (olddb.options.dbVersion == "1.6") then
		olddb.options.minimap_icon = nil
		olddb.options.minimap_icon = {}
		olddb.options.minimap_icon.hide = true
		olddb.options.dbVersion = "1.7"
		addon.Print(string.format(L["Database upgraded to %s"], "1.7"))
	end
	if (olddb.options.dbVersion == "1.7") then
		olddb.options.showCounters = false
		olddb.options.dbVersion = "1.8"
		addon.Print(string.format(L["Database upgraded to %s"], "1.8"))
	end
	if (olddb.options.dbVersion == "1.8") then
		for k, v in pairs(olddb.cooldowns) do
			olddb.cooldowns[k] = nil
		end
		olddb.cooldowns = nil
		olddb.cooldowns = {}
		olddb.options.dbVersion = "1.9"
		addon.Print(string.format(L["Database upgraded to %s"], "1.9"))
		addon.Print(L["This version upgrade required a clear of cooldowns"])
		addon.Print(L["to remove possible corruption. Please rescan your professions."])
	end
end

function addon.stopScanning()
	if (BTC_ScanFrame) then
		BTC_ScanFrame:SetScript("OnUpdate", nil)
	end
end

function addon.scanSkills()
	-- on first load, there's no real good way to check if the player's skills
	-- are available for scanning yet. So just scan every 5 seconds until they are
	if (not BTC_ScanFrame) then
		BTC_ScanFrame = CreateFrame("Frame")
	end

	local totalElapsed = 0
	local checkInterval = 60
	local gotFirst = false

	BTC_ScanFrame:SetScript("OnUpdate", function(self, elapsed)
		totalElapsed = totalElapsed + elapsed
		if (totalElapsed < checkInterval) then
			return
		end
		totalElapsed = 0
		UpdateSkills()
		addon.updateText()
		return
	end)
end

-- on init
function BTC:OnInitialize()
end

function BTC:OnEnable()
	-- Called when the addon is loaded
	self.db = LibStub("AceDB-3.0"):New("BTC_DB", nil, "Default")
	if (not self.db.factionrealm) then
		self.db.factionrealm = {}
	end

	BTC.me = UnitName("player")

	-- do we have any data yet for this toon? If not, init its table
	if not (self.db.factionrealm[BTC.me]) then
		self.db.factionrealm[BTC.me] = {
			primary = {},
			secondary = {},
			class = {},
			languages = {},
			armor = {},
			cooldowns = {},
			options = {
				dbVersion = "1.8",
				debug = false,
				showSecondary = false,
				showText = false,
				showCounters = false,
				hideReminders = false,
				textCooldown = 0,
				tipscale = 1,
				minimap_icon = {
					hide = true,
				},
			}
		}
	else
		-- check if we need to upgrade the db
		BTC_upgradeDB()
	end

	db = self.db.factionrealm[BTC.me]

	addon.Debug("In OnEnable")

	BTC:SetupOptions()

	-- Check current weapons/languages
	addon.scanSkills()
	addon.updateText()

	-- Event listeners
	BTC:RegisterEvent("TRADE_SKILL_SHOW")
	BTC:RegisterEvent("TRADE_SKILL_CLOSE")
	BTC:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	BTC:RegisterEvent("CHAT_MSG_SKILL", UpdateSkills)

	-- libdbicon stuff
	if LDBIcon then
		LDBIcon:Register(ADDON_NAME_SHORT, launcher, db.options.minimap_icon)
	end
end
